Fix RMW bug
[c11tester.git] / action.h
index 7c9146fdda694adfd97c7f932f75822e53a4303c..760e9fb0409143764260103566316525e06ac79c 100644 (file)
--- a/action.h
+++ b/action.h
@@ -5,49 +5,61 @@
 #ifndef __ACTION_H__
 #define __ACTION_H__
 
-#include <list>
 #include <cstddef>
 #include <inttypes.h>
 
 #include "mymemory.h"
 #include "memoryorder.h"
 #include "modeltypes.h"
+#include "mypthread.h"
+#include "classlist.h"
 
-class ClockVector;
-class Thread;
-class Promise;
+namespace cdsc {
+       class mutex;
+}
 
 using std::memory_order;
 using std::memory_order_relaxed;
+using std::memory_order_consume;
 using std::memory_order_acquire;
 using std::memory_order_release;
 using std::memory_order_acq_rel;
 using std::memory_order_seq_cst;
 
-/** Note that this value can be legitimately used by a program, and
- *  hence by iteself does not indicate no value. */
+/**
+ * @brief A recognizable don't-care value for use in the ModelAction::value
+ * field
+ *
+ * Note that this value can be legitimately used by a program, and hence by
+ * iteself does not indicate no value.
+ */
 #define VALUE_NONE 0xdeadbeef
 
-/** A special value to represent a successful trylock */
-#define VALUE_TRYSUCCESS 1
-
-/** A special value to represent a failed trylock */
-#define VALUE_TRYFAILED 0
+/**
+ * @brief The "location" at which a fence occurs
+ *
+ * We need a non-zero memory location to associate with fences, since our hash
+ * tables don't handle NULL-pointer keys. HACK: Hopefully this doesn't collide
+ * with any legitimate memory locations.
+ */
+#define FENCE_LOCATION ((void *)0x7)
 
 /** @brief Represents an action type, identifying one of several types of
  * ModelAction */
 typedef enum action_type {
-       MODEL_FIXUP_RELSEQ,   /**< Special ModelAction: finalize a release
-                              *   sequence */
        THREAD_CREATE,        /**< A thread creation action */
        THREAD_START,         /**< First action in each thread */
        THREAD_YIELD,         /**< A thread yield action */
        THREAD_JOIN,          /**< A thread join action */
        THREAD_FINISH,        /**< A thread completion action */
+       PTHREAD_CREATE,       /**< A pthread creation action */
+       PTHREAD_JOIN,         /**< A pthread join action */
+       
        ATOMIC_UNINIT,        /**< Represents an uninitialized atomic */
        ATOMIC_READ,          /**< An atomic read action */
        ATOMIC_WRITE,         /**< An atomic write action */
        ATOMIC_RMWR,          /**< The read part of an atomic RMW action */
+       ATOMIC_RMWRCAS,          /**< The read part of an atomic RMW action */
        ATOMIC_RMW,           /**< The write part of an atomic RMW action */
        ATOMIC_RMWC,          /**< Convert an atomic RMW action into a READ */
        ATOMIC_INIT,          /**< Initialization of an atomic object (e.g.,
@@ -58,38 +70,46 @@ typedef enum action_type {
        ATOMIC_UNLOCK,        /**< An unlock action */
        ATOMIC_NOTIFY_ONE,    /**< A notify_one action */
        ATOMIC_NOTIFY_ALL,    /**< A notify all action */
-       ATOMIC_WAIT           /**< A wait action */
+       ATOMIC_WAIT,          /**< A wait action */
+       ATOMIC_ANNOTATION,     /**< An annotation action to pass information
+                                                                                                        to a trace analysis */
+       NOOP
 } action_type_t;
 
-/* Forward declaration */
-class Node;
-class ClockVector;
 
 /**
- * The ModelAction class encapsulates an atomic action.
+ * @brief Represents a single atomic action
+ *
+ * A ModelAction is always allocated as non-snapshotting, because it is used in
+ * multiple executions during backtracking. Except for fake uninitialized
+ * (ATOMIC_UNINIT) ModelActions, each action is assigned a unique sequence
+ * number.
  */
 class ModelAction {
 public:
        ModelAction(action_type_t type, memory_order order, void *loc, uint64_t value = VALUE_NONE, Thread *thread = NULL);
+       ModelAction(action_type_t type, memory_order order, void *loc, uint64_t value, int size);
        ~ModelAction();
        void print() const;
 
        thread_id_t get_tid() const { return tid; }
        action_type get_type() const { return type; }
        memory_order get_mo() const { return order; }
+       memory_order get_original_mo() const { return original_order; }
+       void set_mo(memory_order order) { this->order = order; }
        void * get_location() const { return location; }
        modelclock_t get_seq_number() const { return seq_number; }
        uint64_t get_value() const { return value; }
        uint64_t get_reads_from_value() const;
        uint64_t get_write_value() const;
+       uint64_t get_return_value() const;
        const ModelAction * get_reads_from() const { return reads_from; }
-       Promise * get_reads_from_promise() const { return reads_from_promise; }
+       cdsc::mutex * get_mutex() const;
 
        Node * get_node() const;
        void set_node(Node *n) { node = n; }
 
        void set_read_from(const ModelAction *act);
-       void set_read_from_promise(Promise *promise);
 
        /** Store the most recent fence-release from the same thread
         *  @param fence The fence-release that occured prior to this */
@@ -101,7 +121,7 @@ public:
        void set_seq_number(modelclock_t num);
        void set_try_lock(bool obtainedlock);
        bool is_thread_start() const;
-       bool is_relseq_fixup() const;
+       bool is_thread_join() const;
        bool is_mutex_op() const;
        bool is_lock() const;
        bool is_trylock() const;
@@ -115,12 +135,15 @@ public:
        bool is_uninitialized() const;
        bool is_read() const;
        bool is_write() const;
+       bool is_yield() const;
        bool could_be_write() const;
        bool is_rmwr() const;
+       bool is_rmwrcas() const;
        bool is_rmwc() const;
        bool is_rmw() const;
        bool is_fence() const;
        bool is_initialization() const;
+       bool is_annotation() const;
        bool is_relaxed() const;
        bool is_acquire() const;
        bool is_release() const;
@@ -129,7 +152,8 @@ public:
        bool same_thread(const ModelAction *act) const;
        bool is_conflicting_lock(const ModelAction *act) const;
        bool could_synchronize_with(const ModelAction *act) const;
-
+  int getSize() const;
+  
        Thread * get_thread_operand() const;
 
        void create_cv(const ModelAction *parent = NULL);
@@ -149,53 +173,76 @@ public:
        void process_rmw(ModelAction * act);
        void copy_typeandorder(ModelAction * act);
 
-       void set_sleep_flag() { sleep_flag=true; }
-       bool get_sleep_flag() { return sleep_flag; }
        unsigned int hash() const;
 
        bool equals(const ModelAction *x) const { return this == x; }
-       bool equals(const Promise *x) const { return false; }
 
-       bool may_read_from(const ModelAction *write) const;
-       bool may_read_from(const Promise *promise) const;
        MEMALLOC
+
+       void set_value(uint64_t val) { value = val; }
+
+       /* to accomodate pthread create and join */
+       Thread * thread_operand;
+       void set_thread_operand(Thread *th) { thread_operand = th; } 
 private:
 
-       /** Type of action (read, write, thread create, thread yield, thread join) */
+       const char * get_type_str() const;
+       const char * get_mo_str() const;
+
+       /** @brief Type of action (read, write, RMW, fence, thread create, etc.) */
        action_type type;
 
-       /** The memory order for this operation. */
+       /** @brief The memory order for this operation. */
        memory_order order;
 
-       /** A pointer to the memory location for this action. */
+       /** @brief The original memory order parameter for this operation. */
+       memory_order original_order;
+
+       /** @brief A pointer to the memory location for this action. */
        void *location;
 
-       /** The thread id that performed this action. */
+       /** @brief The thread id that performed this action. */
        thread_id_t tid;
 
-       /** The value written (for write or RMW; undefined for read) */
+       /** @brief The value written (for write or RMW; undefined for read) */
        uint64_t value;
 
-       /** The action that this action reads from. Only valid for reads */
+  union {
+  /**
+        * @brief The store that this action reads from
+        *
+        * Only valid for reads
+        */
        const ModelAction *reads_from;
-
-       /** The promise that this action reads from. Only valid for reads */
-       Promise *reads_from_promise;
-
-       /** The last fence release from the same thread */
+    int size;
+  };
+       /** @brief The last fence release from the same thread */
        const ModelAction *last_fence_release;
 
-       /** A back reference to a Node in NodeStack, if this ModelAction is
-        * saved on the NodeStack. */
+       /**
+        * @brief A back reference to a Node in NodeStack
+        *
+        * Only set if this ModelAction is saved on the NodeStack. (A
+        * ModelAction can be thrown away before it ever enters the NodeStack.)
+        */
        Node *node;
 
+       /**
+        * @brief The sequence number of this action
+        *
+        * Except for ATOMIC_UNINIT actions, this number should be unique and
+        * should represent the action's position in the execution order.
+        */
        modelclock_t seq_number;
 
-       /** The clock vector stored with this action; only needed if this
-        * action is a store release? */
+       /**
+        * @brief The clock vector for this operation
+        *
+        * Technically, this is only needed for potentially synchronizing
+        * (e.g., non-relaxed) operations, but it is very handy to have these
+        * vectors for all operations.
+        */
        ClockVector *cv;
-
-       bool sleep_flag;
 };
 
 #endif /* __ACTION_H__ */